﻿// This file contains your Data Connector logic
[Version = "1.0.0"]
section VertexFlow;


[DataSource.Kind="VertexFlow", Publish="VertexFlow.Publish"]
//shared VertexFlow.Contents = Value.ReplaceType(VertexFlowImplementation, VertexFlowType);
shared VertexFlow.Contents = Value.ReplaceType(VertexFlowNavTable, VertexFlowSimpleType);


VertexFlowSimpleType = type function (
    address as (type text meta [
        Documentation.FieldCaption = "Flow server address",
        Documentation.FieldDescription = "Example: https://flow.vertex.fi/",
        Documentation.SampleValues = {"https://flow.vertex.fi"}
    ])
    )
    as binary meta [
        Documentation.Name = "Vertex Flow Contents",
        Documentation.LongDescription = "Return value of the request made to the Vertex Flow API",
        Documentation.Examples = {[
            Description = "Returns the version of the Flow server, database and API",
            Code = "VertexFlow.Contents(""https://flow.vertex.fi"", ""about/version"")",
            Result = "{
                ""core-version"":""20.0.03 (20200130-0002/r47857)"",
                ""core"":""PDM"",
                ""db-version"":""PDM_20_0_03.2001"",
                ""openapi-version"":""1.0.0"",
                ""core-beta"":false}"
        ]}
    ];

VertexFlowType = type function (
    address as (type text meta [
        Documentation.FieldCaption = "Address",
        Documentation.FieldDescription = "Address of the Flow server",
        Documentation.SampleValues = {"https://flow.vertex.fi"}
    ]),
    optional path as (type text meta [
        Documentation.FieldCaption = "Path",
        Documentation.FieldDescription = "End part of the URL path to the service to be requested",
        Documentation.SampleValues = {"about/version", "objects/getSingle"}
    ]),
    optional body as (type text meta [
        Documentation.FieldCaption = "Body",
        Documentation.FieldDescription = "Additional parameters in JSON format, required by certain services",
        Documentation.SampleValues = {"{ ""objCode"" : ""P000004"" }"}
    ])
    )
    as binary meta [
        Documentation.Name = "Vertex Flow Contents",
        Documentation.LongDescription = "Return value of the request made to the Vertex Flow API",
        Documentation.Examples = {[
            Description = "Returns the version of the Flow server, database and API",
            Code = "VertexFlow.Contents(""https://flow.vertex.fi/api/rest/v1/"", ""about/version"")",
            Result = "{
                ""core-version"":""20.0.03 (20200130-0002/r47857)"",
                ""core"":""PDM"",
                ""db-version"":""PDM_20_0_03.2001"",
                ""openapi-version"":""1.0.0"",
                ""core-beta"":false}"
        ]}
    ];

DirectAPICallHandlerType = type function (
    servicePath as (type text meta [
        Documentation.FieldCaption = "Service path",
        Documentation.FieldDescription = "End part of the URL path to the service to be requested",
        Documentation.SampleValues = {"about/version", "objects/getSingle"}
    ]),
    optional jsonParameters as (type text meta [
        Documentation.FieldCaption = "JSON parameters",
        Documentation.FieldDescription = "Additional parameters in JSON format, required by certain services",
        Documentation.SampleValues = {"{ ""objCode"" : ""P000004"" }"}
    ])
    )
    as binary meta [
        Documentation.Name = "Vertex Flow API direct call result",
        Documentation.LongDescription = "Return value of the request made to the Vertex Flow API",
        Documentation.Examples = {[
            Description = "Returns the version of the Flow server, database and API",
            Code = "DirectAPICallHandler(""about/version"")",
            Result = "{
                ""core-version"":""20.0.03 (20200130-0002/r47857)"",
                ""core"":""PDM"",
                ""db-version"":""PDM_20_0_03.2001"",
                ""openapi-version"":""1.0.0"",
                ""core-beta"":false}"
        ]}
    ];

GetObjectConnectionsRecursiveHandlerType = type function (
    objCode as (type text meta [
        Documentation.FieldCaption = "Object code",
        Documentation.FieldDescription = "Object code of the parent object",
        Documentation.SampleValues = {"P000025"}
    ]),
    hierarchyLevels as (type number meta [
        Documentation.FieldCaption = "Hierarchy levels",
        Documentation.FieldDescription = "Depth of the hierarchy to get. 0 = only parent, 1 = only direct connections, etc.",
        Documentation.SampleValues = {0, 1, 2}
    ]),
    optional removeDuplicates as (type logical meta [
        Documentation.FieldCaption = "Remove duplicates",
        Documentation.FieldDescription = "Defaults to true. If true, removes duplicate objects from the result.",
        Documentation.SampleValues = {true, false}
    ])
    )
    as table meta [
        Documentation.Name = "Direct and indirect object connections",
        Documentation.LongDescription = "Return value of the request made to the Vertex Flow API",
        Documentation.Examples = {[
            Description = "Two hierarchy levels down",
            Code = "GetObjectConnectionsRecursiveHandler(""P000025"", 2)",
            Result = "Table containing (Parent), (direct connections of parent) and (direct connections of direct connections of parent)"
        ]}
    ];
    

//VertexFlowImplementation = (
DirectAPICall = (
    address as text,        // E.g. "https://flow.vertex.fi"
    optional path as text,  // E.g. "about/version", optional to work around Power BI caching expired tokens (defaults to "about/version")
    optional body as text   // {json}
    ) =>
    let
        // If path is null, it defaults to "about/version"
        service_path = if(path = null) then
                "about/version"
            else
                path,

        // Drop possible trailing "/" character from the address
        sanitizedAddress = SanitizeAddress(address),

        // Check if the request requires authorization and create the correct request
        result = if (service_path = "about/version") then
                Web.Contents(sanitizedAddress & "/api/rest/v1/" & service_path)
            else
                // The function sanitizes the address on its own
                GetRequestThroughAuthorization(address, service_path, body),

        jsonResult = if (path = "users/hello")
            then
                result
            else
                Json.Document(result, 65001)
    in
        jsonResult;

GetRequestThroughAuthorization = (
    address as text,        // E.g. "https://flow.vertex.fi/api/rest/v1/"
    path as text,           // E.g. "objects/getSingle"
    optional body as text,  // E.g. "{ "objCode" : "P000004" }
    optional queryParameters as list
    )
    as binary =>
    let
        // Drop possible trailing "/" character from the address
        sanitizedAddress = SanitizeAddress(address),

        // Get current Flow credentials stored in Power BI
        current_credential = Extension.CurrentCredential(),

        // Login to Flow using the credentials, store the authorization token
        token = GetAuthorizationToken(sanitizedAddress, current_credential[Username], current_credential[Password]),
        auth = "Bearer " & token,

        parsedQueryParameters = if (queryParameters <> null and List.Count(queryParameters) > 0)
            then
                ParseQueryParameters(queryParameters)
            else
                "",

        // Call Flow OpenAPI service using the authorization token (if necessary)
        result = if (body = null) then
                Web.Contents(sanitizedAddress & "/api/rest/v1/" & path & parsedQueryParameters, [Headers=[#"Authorization"=auth]])
            else
                Web.Contents(sanitizedAddress & "/api/rest/v1/" & path & parsedQueryParameters,
                            [Headers=[
                                #"Authorization"=auth,
                                #"Content-Type"="application/json"],
                             Content=Text.ToBinary(body)])
    in
        result;

GetAuthorizationToken = (address as text, username as text, password as text) =>
    let
        login_body = "{""username"":""" & username & """,""password"":""" & password & """}",

        // Power BI likes to cache results, which means that the authorization token
        // is not always re-evaluated when necessary, but rather read from the cache.
        // We can prevent that by using a dummy variable
        disable_cache = "?disable_cache=" & timestamp,
        timestamp = DateTimeZone.ToText(DateTimeZone.LocalNow()),

        login_url = address & "/api/rest/v1/" & "users/login" & disable_cache,

        // Call the login service using the given credentials and
        // store the response
        response = Csv.Document(
            Web.Contents(
                login_url,
                [Headers=
                    [#"Content-Type"="application/json"],
                    Content=Text.ToBinary(login_body)
                ]
            ),
            [Delimiter=";", Columns=1, Encoding=1252, QuoteStyle=QuoteStyle.None]),

        // Read the authorization token from the response
        token = List.First(response[Column1])
    in
        // Return the authorization token
        token;

shared currentAddress = "";

ParseQueryParameters = (queryParameters as list) =>
    let
        joinKeysAndValues   = List.Transform(queryParameters, each Text.Combine(_, "=")),
        joinParameters      = Text.Combine(joinKeysAndValues, "&"),
        parsedQueryParameters = "?" & joinParameters
    in
        parsedQueryParameters;

SanitizeAddress = (address as text) =>
    let
        sanitizedAddress = if (Text.EndsWith(address, "/")) // Remove trailing "/"
            then
              Text.Range(address, 0, (Text.Length(address) - 1))
            else
              address
    in
        sanitizedAddress;

GetJsonThroughAuthorization = (
    address as text,        // E.g. "https://flow.vertex.fi/api/rest/v1/"
    path as text,           // E.g. "objects/getSingle"
    optional body as text,  // E.g. "{ "objCode" : "P000004" }
    optional queryParameters as list
    )
    as any =>
    let
        response = GetRequestThroughAuthorization(
            address,
            path,
            body,
            queryParameters),

        metadata = Value.Metadata(response),
        status = metadata[Response.Status],

        json = if (status = 200)
            then
                Json.Document(response, 65001)
            else
                if (status = 204) // No content
                    then
                        ""
                    else
                        Json.Document(response, 65001)
    in
        json;

GetTableThroughAuthorization = (
    address as text,        // E.g. "https://flow.vertex.fi/api/rest/v1/"
    path as text,           // E.g. "objects/getSingle"
    optional body as text,  // E.g. "{ "objCode" : "P000004" }
    optional queryParameters as list
    )
    as any =>
    let
        json = GetJsonThroughAuthorization(
            address,
            path,
            body,
            queryParameters),

        table = if (json <> "")
            then
                if (json is list)
                    then
                        Table.FromRecords(json) // List of records to table
                    else
                        Record.ToTable(json) // Single record to table
            else
                #table({},{}) // Empty json to empty table
    in
        table;

VertexFlowNavTable = (
    address as text/*,        // E.g. "https://flow.vertex.fi/api/rest/v1/"
    optional path as text,  // E.g. "about/version", optional to work around Power BI caching expired tokens (defaults to "about/version")
    optional body as text   // {json}*/
    ) as table =>
    let
        
        currentAddress = Table.Buffer(address),

        /*********************************************************************/
        // Local helper functions that see "address" in this scope

        AboutSettingsHandler = (key as text) =>
            let
                result = GetSettings(address, key)
            in
                result,
        
        ObjectsConnectionsHandler = (objType as text, objCode as text, optional getStateHistory as logical) =>
            let
                result = GetObjectConnectionsDirect(address, objType, objCode, getStateHistory)
            in
                result,
        
        ObjectsFindInConnectionsHandler = (
            objType as text,
            objCode as text,
            optional filterBaseClass as text,
            optional filterTypeCode as text,
            optional filterTypeName as text,
            optional filterObjCode as text,
            optional filterRevision as text,
            optional filterIsOldRevision as text,
            optional filterCheckedOutByUserId as text,
            optional filterName as text,
            optional filterDescription as text,
            optional filterComment as text,
            optional filterAuthor as text,
            optional filterAuthorId as text,
            optional filterCreator as text,
            optional filterCreatorId as text,
            optional filterOwner as text,
            optional filterOwnerId as text,
            optional filterPublicity as text,
            optional filterPublicityName as text,
            optional filterClassification as text,
            optional filterClassificationName as text,
            optional filterStatus as text,
            optional filterStatusName as text,
            //optional filterCreationTime as text,
            //optional filterModifiedTime as text,
            optional filterExtAttributeKey0 as text,
            optional filterExtAttributeValue0 as text,
            optional filterExtAttributeKey1 as text,
            optional filterExtAttributeValue1 as text,
            optional filterExtAttributeKey2 as text,
            optional filterExtAttributeValue2 as text,
            optional filterExtAttributeKey3 as text,
            optional filterExtAttributeValue3 as text,
            optional filterExtAttributeKey4 as text,
            optional filterExtAttributeValue4 as text,
            optional filterExtAttributeKey5 as text,
            optional filterExtAttributeValue5 as text,
            optional filterExtAttributeKey6 as text,
            optional filterExtAttributeValue6 as text,
            optional filterExtAttributeKey7 as text,
            optional filterExtAttributeValue7 as text,
            optional filterExtAttributeKey8 as text,
            optional filterExtAttributeValue8 as text,
            optional filterExtAttributeKey9 as text,
            optional filterExtAttributeValue9 as text,
            optional getStateHistory as logical
            ) =>
            let
                result = ObjectsFindInConnections(
                    address,
                    objType,
                    objCode,
                    filterBaseClass,
                    filterTypeCode,
                    filterTypeName,
                    filterObjCode,
                    filterRevision,
                    filterIsOldRevision,
                    filterCheckedOutByUserId,
                    filterName,
                    filterDescription,
                    filterComment,
                    filterAuthor,
                    filterAuthorId,
                    filterCreator,
                    filterCreatorId,
                    filterOwner,
                    filterOwnerId,
                    filterPublicity,
                    filterPublicityName,
                    filterClassification,
                    filterClassificationName,
                    filterStatus,
                    filterStatusName,
                    //filterCreationTime,
                    //filterModifiedTime,
                    filterExtAttributeKey0,
                    filterExtAttributeValue0,
                    filterExtAttributeKey1,
                    filterExtAttributeValue1,
                    filterExtAttributeKey2,
                    filterExtAttributeValue2,
                    filterExtAttributeKey3,
                    filterExtAttributeValue3,
                    filterExtAttributeKey4,
                    filterExtAttributeValue4,
                    filterExtAttributeKey5,
                    filterExtAttributeValue5,
                    filterExtAttributeKey6,
                    filterExtAttributeValue6,
                    filterExtAttributeKey7,
                    filterExtAttributeValue7,
                    filterExtAttributeKey8,
                    filterExtAttributeValue8,
                    filterExtAttributeKey9,
                    filterExtAttributeValue9,
                    getStateHistory)
            in
                result,

        ObjectsFindHandler = (jsonParameters as text, optional getStateHistory as logical) =>
            let
                getStateHistoryNotNull = if (getStateHistory = null)
                  then
                    false
                  else
                    getStateHistory,
                result = ObjectsFind(address, jsonParameters, {{"get-state-history", Logical.ToText(getStateHistoryNotNull)}})
            in
                result,

        ObjectsFindByParametersHandler = (
            optional baseClass as text,
            optional typeCode as text,
            //optional typeName as text, // Not yet implemented in back-end SQL search
            optional objCode as text,
            optional revision as text,
            //optional isOldRevision as text, // Not yet implemented in back-end SQL search
            //optional checkedOutByUserId as text, // Not yet implemented in back-end SQL search
            optional name as text,
            optional description as text,
            //optional comment as text, // Not yet implemented in back-end SQL search
            //optional author as text, // Not yet implemented in back-end SQL search
            optional authorId as text,
            //optional creator as text, // Not yet implemented in back-end SQL search
            optional creatorId as text,
            //optional owner as text, // Not yet implemented in back-end SQL search
            //optional ownerId as text, // Not yet implemented in back-end SQL search
            //optional publicity as text, // Not yet implemented in back-end SQL search
            //optional publicityName as text, // Not yet implemented in back-end SQL search
            optional classification as text,
            //optional classificationName as text, // Not yet implemented in back-end SQL search
            optional status as text,
            //optional statusName as text, // Not yet implemented in back-end SQL search
            optional extAttributeKey0 as text,
            optional extAttributeValue0 as text,
            optional extAttributeKey1 as text,
            optional extAttributeValue1 as text,
            optional extAttributeKey2 as text,
            optional extAttributeValue2 as text,
            optional extAttributeKey3 as text,
            optional extAttributeValue3 as text,
            optional extAttributeKey4 as text,
            optional extAttributeValue4 as text,
            optional extAttributeKey5 as text,
            optional extAttributeValue5 as text,
            optional extAttributeKey6 as text,
            optional extAttributeValue6 as text,
            optional extAttributeKey7 as text,
            optional extAttributeValue7 as text,
            optional extAttributeKey8 as text,
            optional extAttributeValue8 as text,
            optional extAttributeKey9 as text,
            optional extAttributeValue9 as text,
            optional getStateHistory as logical
            ) =>
            let
                result = ObjectsFindByParameters(
                    address,
                    baseClass,
                    typeCode,
                    //typeName,
                    objCode,
                    revision,
                    //isOldRevision,
                    //checkedOutByUserId,
                    name,
                    description,
                    //comment,
                    //author,
                    authorId,
                    //creator,
                    creatorId,
                    //owner,
                    //ownerId,
                    //publicity,
                    //publicityName,
                    classification,
                    //classificationName,
                    status,
                    //statusName,
                    extAttributeKey0,
                    extAttributeValue0,
                    extAttributeKey1,
                    extAttributeValue1,
                    extAttributeKey2,
                    extAttributeValue2,
                    extAttributeKey3,
                    extAttributeValue3,
                    extAttributeKey4,
                    extAttributeValue4,
                    extAttributeKey5,
                    extAttributeValue5,
                    extAttributeKey6,
                    extAttributeValue6,
                    extAttributeKey7,
                    extAttributeValue7,
                    extAttributeKey8,
                    extAttributeValue8,
                    extAttributeKey9,
                    extAttributeValue9,
                    getStateHistory)
            in
                result,

        ObjectsGetSingleHandler = (typeCode as text, objCode as text, optional getStateHistory as logical) =>
            let
                result = ObjectsGetSingleByTypeAndObjCode(address, typeCode, objCode, getStateHistory)
            in
                result,

        ObjectsStateHistoryHandler = (typeCode as text, objCode as text) =>
            let
                result = ObjectsGetStateHistory(address, typeCode, objCode)
            in
                result,

        ProjectsBOMHandler = (objCode as text) =>
            let
                result = GetProjectBOM(address, objCode)
            in
                result,

        ProjectsFindHandler = (
            projectCode as text,
            optional baseClass as text,
            optional typeCode as text,
            optional typeName as text,
            optional objCode as text,
            optional revision as text,
            optional isOldRevision as text,
            optional checkedOutByUserId as text,
            optional name as text,
            optional description as text,
            optional comment as text,
            optional author as text,
            optional authorId as text,
            optional creator as text,
            optional creatorId as text,
            optional owner as text,
            optional ownerId as text,
            optional publicity as text,
            optional publicityName as text,
            optional classification as text,
            optional classificationName as text,
            optional status as text,
            optional statusName as text,
            optional extAttributeKey0 as text,
            optional extAttributeValue0 as text,
            optional extAttributeKey1 as text,
            optional extAttributeValue1 as text,
            optional extAttributeKey2 as text,
            optional extAttributeValue2 as text,
            optional extAttributeKey3 as text,
            optional extAttributeValue3 as text,
            optional extAttributeKey4 as text,
            optional extAttributeValue4 as text,
            optional extAttributeKey5 as text,
            optional extAttributeValue5 as text,
            optional extAttributeKey6 as text,
            optional extAttributeValue6 as text,
            optional extAttributeKey7 as text,
            optional extAttributeValue7 as text,
            optional extAttributeKey8 as text,
            optional extAttributeValue8 as text,
            optional extAttributeKey9 as text,
            optional extAttributeValue9 as text,
            optional getStateHistory as logical
            ) =>
              let
                result = FindUnderProjectByParameters(
                    address,
                    projectCode,
                    baseClass,
                    typeCode,
                    typeName,
                    objCode,
                    revision,
                    isOldRevision,
                    checkedOutByUserId,
                    name,
                    description,
                    comment,
                    author,
                    authorId,
                    creator,
                    creatorId,
                    owner,
                    ownerId,
                    publicity,
                    publicityName,
                    classification,
                    classificationName,
                    status,
                    statusName,
                    extAttributeKey0,
                    extAttributeValue0,
                    extAttributeKey1,
                    extAttributeValue1,
                    extAttributeKey2,
                    extAttributeValue2,
                    extAttributeKey3,
                    extAttributeValue3,
                    extAttributeKey4,
                    extAttributeValue4,
                    extAttributeKey5,
                    extAttributeValue5,
                    extAttributeKey6,
                    extAttributeValue6,
                    extAttributeKey7,
                    extAttributeValue7,
                    extAttributeKey8,
                    extAttributeValue8,
                    extAttributeKey9,
                    extAttributeValue9,
                    getStateHistory)
            in
                result,

        GetObjectConnectionsRecursiveHandler = (objCode as text, hierarchyLevels as number, optional removeDuplicates as logical) =>
            let
                shouldRemoveDuplicates = if (removeDuplicates = null)
                    then
                        true    // Defaults to true
                    else
                        removeDuplicates,

                result = GetObjectConnectionsRecursiveHelper(address, objCode, hierarchyLevels, shouldRemoveDuplicates)
            in
                result,

        DirectAPICallHandler = (servicePath as text, optional jsonParameters as text) =>
            let
                result = DirectAPICall(address, servicePath, jsonParameters)
            in
                result,
        
        GetFileListHandler = (
            baseClass as text,
            typeCode  as text,
            objCode   as text,
            revision  as text) =>
            let
                result = GetFileList(address, baseClass, typeCode, objCode, revision)
            in
                result,

        GetUserListHandler = () =>
            let
                result = GetUsersList(address)
            in
                result,

        /*********************************************************************/

        /*************
        * About
        *************/

        aboutTable = #table(
                {"Name",            "Key",                  "Data",                         "ItemKind", "ItemName", "IsLeaf"},
            {
                {"Settings",        "about/settings",       Value.ReplaceType(
                                                                AboutSettingsHandler,
                                                                AboutSettingsHandlerType),  "Table",    "Function", true},
                {"Version",         "about/version",        GetVersion(address),            "Table",    "Table",    true}
            }),

        aboutNavTable = Table.ToNavigationTableNoDelay(
            aboutTable, {"Key"}, "Name", "Data", "ItemKind", "ItemName", "IsLeaf"),
            

        /*************
        * Objects
        *************/

        objectsFunctionsSubTable = #table (
                {"Name",                            "Key",                          "Data",                                     "ItemKind", "ItemName", "IsLeaf"},
            {
                {"Get single",                      "objects/getSingle",            Value.ReplaceType(
                                                                                        ObjectsGetSingleHandler,
                                                                                        TypeAndObjCodeToTableHandlerWithGetStateHistoryType),
                                                                                                                                "Table",    "Function", true},
                {"Get all connections",             "objects/connections",          Value.ReplaceType(
                                                                                        ObjectsConnectionsHandler,
                                                                                        ObjectsConnectionsHandlerType),         "Table",    "Function", true},
                {"Find in connections",             "objects/findInConnections",    Value.ReplaceType(
                                                                                        ObjectsFindInConnectionsHandler,
                                                                                        ObjectsFindInConnectionsHandlerType),   "Table",    "Function", true},
                {"Find by JSON",                    "objects/findJSON",             Value.ReplaceType(
                                                                                        ObjectsFindHandler,
                                                                                        ObjectsFindHandlerType),                "Table",    "Function", true},
                {"Find by parameters",              "objects/findParams",           Value.ReplaceType(
                                                                                        ObjectsFindByParametersHandler,
                                                                                        ObjectsFindByParametersHandlerType),    "Table",    "Function", true},
                {"Get state history",               "objects/stateHistory",         Value.ReplaceType(
                                                                                        ObjectsStateHistoryHandler,
                                                                                        TypeAndObjCodeToTableHandlerType),      "Table",    "Function", true}
            }),

        objectsFunctionsSubNavTable = Table.ToNavigationTableNoDelay(
            objectsFunctionsSubTable, {"Key"}, "Name", "Data", "ItemKind", "ItemName", "IsLeaf"),

        objectsBrowseSubTable = #table (
                {"Name",                "Key",                          "Data",                         "ItemKind", "ItemName", "IsLeaf"},
            {
                {"Classifications",     "objects/classifications",      GetClassifications(address),    "Table",    "",         true}
            }),

        objectsBrowseSubNavTable = Table.ToNavigationTable(
            objectsBrowseSubTable, {"Key"}, "Name", "Data", "ItemKind", "ItemName", "IsLeaf"),

        objectsTable = #table(
                {"Name",            "Key",                  "Data",                         "ItemKind", "ItemName", "IsLeaf"},
            {
                
                {"Browse",          "objectsBrowse",        objectsBrowseSubNavTable,       "Folder",   "",         false},
                {"Functions",       "objectsFunctions",     objectsFunctionsSubNavTable,    "Folder",   "",         false}
                
            }
        ),
        objectsNavTable = Table.ToNavigationTable(
            objectsTable, {"Key"}, "Name", "Data", "ItemKind", "ItemName", "IsLeaf"),


        /*************
        * Files
        *************/
        
        filesTable = #table (
                {"Name",                    "Key",                "Data",                       "ItemKind", "ItemName", "IsLeaf"},
            {
                {"Get file list of object", "getFileList",        Value.ReplaceType(
                                                                      GetFileListHandler,
                                                                      GetFileListHandlerType),  "Table",    "Function", true}
            }),

        filesNavTable = Table.ToNavigationTableNoDelay(
            filesTable, {"Key"}, "Name", "Data", "ItemKind", "ItemName", "IsLeaf"),

        /*************
        * Functions
        *************/

        functionSubTable = #table(
                {"Name",                    "Key",              "Data",                                         "ItemKind", "ItemName", "IsLeaf"},
            {
                {"Call API directly",           "directAPIcall",    Value.ReplaceType(
                                                                        DirectAPICallHandler,
                                                                        DirectAPICallHandlerType),                  "Function", "Function", true}
            }
        ),

        functionNavTable = Table.ToNavigationTableNoDelay(
            functionSubTable, {"Key"}, "Name", "Data", "ItemKind", "ItemName", "IsLeaf"),

        /*************
        * Projects
        *************/

        projectsFunctionsSubTable = #table (
                {"Name",            "Key",                  "Data",                             "ItemKind", "ItemName", "IsLeaf"},
            {
                {"Get BOM",             "projects/bom",         Value.ReplaceType(
                                                                ProjectsBOMHandler,
                                                                ProjectsBOMHandlerType),        "Table",    "Function", true},
                {"Find under project",  "projects/find",        Value.ReplaceType(
                                                                ProjectsFindHandler,
                                                                ProjectsFindHandlerType),       "Table",    "Function", true}
            }),

        projectsFunctionsSubNavTable = Table.ToNavigationTableNoDelay(
            projectsFunctionsSubTable, {"Key"}, "Name", "Data", "ItemKind", "ItemName", "IsLeaf"),

        projectsBrowseSubTable = #table (
                {"Name",            "Key",                  "Data",                             "ItemKind", "ItemName", "IsLeaf"},
            {
                {"All projects",        "projects",             GetProjects(address),           "Table",    "",         true},
                {"Project hierarchy",   "projectsRecursive",    GetProjectsRecursive(address),  "Cube",     "",         true}
            }),

        projectsBrowseSubNavTable = Table.ToNavigationTable(
            projectsBrowseSubTable, {"Key"}, "Name", "Data", "ItemKind", "ItemName", "IsLeaf"),

        projectsTable = #table (
                {"Name",            "Key",                  "Data",                         "ItemKind", "ItemName", "IsLeaf"},
            {
                {"Browse",          "projectsBrowse",       projectsBrowseSubNavTable,      "Folder",   "",         false},
                {"Functions",       "projectsFunctions",    projectsFunctionsSubNavTable,   "Folder",   "",         false}
            }),

        projectsNavTable = Table.ToNavigationTableNoDelay(
            projectsTable, {"Key"}, "Name", "Data", "ItemKind", "ItemName", "IsLeaf"),


        /*************
        * Users
        *************/

        usersTable = #table (
                {"Name",                    "Key",                "Data",                     "ItemKind", "ItemName", "IsLeaf"},
            {
                {"Get user list",           "getUserList",        Value.ReplaceType(
                                                                      GetUserListHandler,
                                                                      NoParametersType),      "Table",    "Function", true}
            }),

        usersNavTable = Table.ToNavigationTableNoDelay(
            usersTable, {"Key"}, "Name", "Data", "ItemKind", "ItemName", "IsLeaf"),

        rootTable = #table(
                {"Name",            "Key",          "Data",                 "ItemKind", "ItemName", "IsLeaf"},
            {
                {"About",           "about",        aboutNavTable,          "Folder",   "Folder",   false},
                {"Files",           "files",        filesNavTable,          "Folder",   "Folder",   false},
                {"Functions",       "functions",    functionNavTable,       "Folder",   "Folder",   false},
                {"Objects",         "objects",      objectsNavTable,        "Folder",   "Folder",   false},
                {"Projects",        "projects",     projectsNavTable,       "Folder",   "Folder",   false},
                {"Users",           "users",        usersNavTable,          "Folder",   "Folder",   false}
                
            }
        ),

        rootNavTable = Table.ToNavigationTable/*NoDelay*/(
            rootTable, {"Key"}, "Name", "Data", "ItemKind", "ItemName", "IsLeaf")
    in
        rootNavTable;

/************************************
* Miscellaneous
************************************/

NoParametersType = type function ()
    as any meta [];

/************************************
* About
************************************/

AboutSettingsHandlerType = type function (
    key as (type text meta [
        Documentation.FieldCaption      = "Settings key",
        Documentation.FieldDescription  = "Name of the settings key",
        Documentation.SampleValues      = {""}
        ])
    )
    as table meta [
        Documentation.Name = "Application setting keys",
        Documentation.LongDescription = "Value of the application specific settings key"];

GetSettings = (address as text, key as text) as table =>
    let
        table = GetTableThroughAuthorization(address, "about/settings/" & key)
    in
        table;


GetVersion = (address as text) as table =>
    let
        response = DirectAPICall(address, "about/version"),
        responseTable = Record.ToTable(response)
    in
        responseTable;


/************************************
* Projects
************************************/

ProjectsBOMHandlerType = type function (
    objCode as (type text meta [
        Documentation.FieldCaption      = "Object code",
        Documentation.FieldDescription  = "Object code of the project",
        Documentation.SampleValues      = {"P000004"}
        ]))
    as table meta [];

GetProjectBOM = (address as text, objCode as text) as table =>
    let
        table = GetTableThroughAuthorization(
            address,
            "projects/" & objCode & "/bom")
    in
        table;

ProjectsFindHandlerType = type function (
    projectCode as (type text meta [
        Documentation.FieldCaption      = "Project code",
        Documentation.FieldDescription  = "Object code of the project",
        Documentation.SampleValues      = {"P000004"}
        ]),
    optional baseClass as (type text meta [
        Documentation.FieldCaption      = "Base class",
        Documentation.FieldDescription  = "Base class of the object",
        Documentation.SampleValues      = {"doc"}
        ]),
    optional typeCode as (type text meta [
        Documentation.FieldCaption      = "Type (code)",
        Documentation.FieldDescription  = "Type of the object (type code)",
        Documentation.SampleValues      = {"model"}
        ]),
    optional typeName as (type text meta [
        Documentation.FieldCaption      = "Type",
        Documentation.FieldDescription  = "Type of the object (type name)",
        Documentation.SampleValues      = {"Model (Vertex)"}
        ]),
    optional objCode as (type text meta [
        Documentation.FieldCaption      = "Object code",
        Documentation.FieldDescription  = "Object code of the object",
        Documentation.SampleValues      = {"P000004"}
        ]),
    optional revision as (type text meta [
        Documentation.FieldCaption      = "Revision",
        Documentation.FieldDescription  = "Revision of the object",
        Documentation.SampleValues      = {"0"}
        ]),
    optional isOldRevision as (type text meta [
        Documentation.FieldCaption      = "Is old revision",
        Documentation.FieldDescription  = "False if this is the newest revision of the object",
        Documentation.SampleValues      = {"false"}
        ]),
    optional checkedOutByUserId as (type text meta [
        Documentation.FieldCaption      = "Checked out by user ID",
        Documentation.FieldDescription  = "The object is checked out by a specific user",
        Documentation.SampleValues      = {""}
        ]),
    optional name as (type text meta [
        Documentation.FieldCaption      = "Name",
        Documentation.FieldDescription  = "Name of the object",
        Documentation.SampleValues      = {""}
        ]),
    optional description as (type text meta [
        Documentation.FieldCaption      = "Description",
        Documentation.FieldDescription  = "Description of the object",
        Documentation.SampleValues      = {""}
        ]),
    optional comment as (type text meta [
        Documentation.FieldCaption      = "Comment",
        Documentation.FieldDescription  = "Comment of the object",
        Documentation.SampleValues      = {""}
        ]),
    optional author as (type text meta [
        Documentation.FieldCaption      = "Author",
        Documentation.FieldDescription  = "Author of the object",
        Documentation.SampleValues      = {""}
        ]),
    optional authorId as (type text meta [
        Documentation.FieldCaption      = "Author ID",
        Documentation.FieldDescription  = "ID of the author of the object",
        Documentation.SampleValues      = {""}
        ]),
    optional creator as (type text meta [
        Documentation.FieldCaption      = "Creator",
        Documentation.FieldDescription  = "Creator of the object",
        Documentation.SampleValues      = {""}
        ]),
    optional creatorId as (type text meta [
        Documentation.FieldCaption      = "Creator ID",
        Documentation.FieldDescription  = "ID of the creator of the object",
        Documentation.SampleValues      = {""}
        ]),
    optional owner as (type text meta [
        Documentation.FieldCaption      = "Owner",
        Documentation.FieldDescription  = "Owner of the object",
        Documentation.SampleValues      = {""}
        ]),
    optional ownerId as (type text meta [
        Documentation.FieldCaption      = "Owner ID",
        Documentation.FieldDescription  = "ID of the owner of the object",
        Documentation.SampleValues      = {""}
        ]),
    optional publicity as (type text meta [
        Documentation.FieldCaption      = "Publicity (code)",
        Documentation.FieldDescription  = "Publicity of the object (publicity code)",
        Documentation.SampleValues      = {""}
        ]),
    optional publicityName as (type text meta [
        Documentation.FieldCaption      = "Publicity",
        Documentation.FieldDescription  = "Publicity of the object (publicity name)",
        Documentation.SampleValues      = {""}
        ]),
    optional classification as (type text meta [
        Documentation.FieldCaption      = "Classification (code)",
        Documentation.FieldDescription  = "Classification of the object (classification code)",
        Documentation.SampleValues      = {""}
        ]),
    optional classificationName as (type text meta [
        Documentation.FieldCaption      = "Classification",
        Documentation.FieldDescription  = "Classification of the object (classification code)",
        Documentation.SampleValues      = {""}
        ]),
    optional status as (type text meta [
        Documentation.FieldCaption      = "Status (code)",
        Documentation.FieldDescription  = "Status of the object",
        Documentation.SampleValues      = {"DRAFT"}
        ]),
    optional statusName as (type text meta [
        Documentation.FieldCaption      = "Status",
        Documentation.FieldDescription  = "Status of the object (status name)",
        Documentation.SampleValues      = {"Draft"}
        ]),
    // Notice off-by-one (programmers start from 0, people start from 1)
    optional extAttributeKey0 as (type text meta [
        Documentation.FieldCaption      = "Extra attribute Key 1",
        Documentation.FieldDescription  = "Extra attribute key of the object",
        Documentation.SampleValues      = {"c_customExtraAttribute"}
        ]),
    optional extAttributeValue0 as (type text meta [
        Documentation.FieldCaption      = "Extra attribute Value 1",
        Documentation.FieldDescription  = "Value of the extra attribute key 1",
        Documentation.SampleValues      = {""}
        ]),
        optional extAttributeKey1 as (type text meta [
        Documentation.FieldCaption      = "Extra attribute Key 2",
        Documentation.FieldDescription  = "Extra attribute key of the object",
        Documentation.SampleValues      = {"c_customExtraAttribute"}
        ]),
    optional extAttributeValue1 as (type text meta [
        Documentation.FieldCaption      = "Extra attribute Value 2",
        Documentation.FieldDescription  = "Value of the extra attribute key 2",
        Documentation.SampleValues      = {""}
        ]),
        optional extAttributeKey2 as (type text meta [
        Documentation.FieldCaption      = "Extra attribute Key 3",
        Documentation.FieldDescription  = "Extra attribute key of the object",
        Documentation.SampleValues      = {"c_customExtraAttribute"}
        ]),
    optional extAttributeValue2 as (type text meta [
        Documentation.FieldCaption      = "Extra attribute Value 3",
        Documentation.FieldDescription  = "Value of the extra attribute key 3",
        Documentation.SampleValues      = {""}
        ]),
        optional extAttributeKey3 as (type text meta [
        Documentation.FieldCaption      = "Extra attribute Key 4",
        Documentation.FieldDescription  = "Extra attribute key of the object",
        Documentation.SampleValues      = {"c_customExtraAttribute"}
        ]),
    optional extAttributeValue3 as (type text meta [
        Documentation.FieldCaption      = "Extra attribute Value 4",
        Documentation.FieldDescription  = "Value of the extra attribute key 4",
        Documentation.SampleValues      = {""}
        ]),
        optional extAttributeKey4 as (type text meta [
        Documentation.FieldCaption      = "Extra attribute Key 5",
        Documentation.FieldDescription  = "Extra attribute key of the object",
        Documentation.SampleValues      = {"c_customExtraAttribute"}
        ]),
    optional extAttributeValue4 as (type text meta [
        Documentation.FieldCaption      = "Extra attribute Value 5",
        Documentation.FieldDescription  = "Value of the extra attribute key 5",
        Documentation.SampleValues      = {""}
        ]),
        optional extAttributeKey5 as (type text meta [
        Documentation.FieldCaption      = "Extra attribute Key 6",
        Documentation.FieldDescription  = "Extra attribute key of the object",
        Documentation.SampleValues      = {"c_customExtraAttribute"}
        ]),
    optional extAttributeValue5 as (type text meta [
        Documentation.FieldCaption      = "Extra attribute Value 6",
        Documentation.FieldDescription  = "Value of the extra attribute key 6",
        Documentation.SampleValues      = {""}
        ]),
        optional extAttributeKey6 as (type text meta [
        Documentation.FieldCaption      = "Extra attribute Key 7",
        Documentation.FieldDescription  = "Extra attribute key of the object",
        Documentation.SampleValues      = {"c_customExtraAttribute"}
        ]),
    optional extAttributeValue6 as (type text meta [
        Documentation.FieldCaption      = "Extra attribute Value 7",
        Documentation.FieldDescription  = "Value of the extra attribute key 7",
        Documentation.SampleValues      = {""}
        ]),
        optional extAttributeKey7 as (type text meta [
        Documentation.FieldCaption      = "Extra attribute Key 8",
        Documentation.FieldDescription  = "Extra attribute key of the object",
        Documentation.SampleValues      = {"c_customExtraAttribute"}
        ]),
    optional extAttributeValue7 as (type text meta [
        Documentation.FieldCaption      = "Extra attribute Value 8",
        Documentation.FieldDescription  = "Value of the extra attribute key 8",
        Documentation.SampleValues      = {""}
        ]),
        optional extAttributeKey8 as (type text meta [
        Documentation.FieldCaption      = "Extra attribute Key 9",
        Documentation.FieldDescription  = "Extra attribute key of the object",
        Documentation.SampleValues      = {"c_customExtraAttribute"}
        ]),
    optional extAttributeValue8 as (type text meta [
        Documentation.FieldCaption      = "Extra attribute Value 9",
        Documentation.FieldDescription  = "Value of the extra attribute key 9",
        Documentation.SampleValues      = {""}
        ]),
        optional extAttributeKey9 as (type text meta [
        Documentation.FieldCaption      = "Extra attribute Key 10",
        Documentation.FieldDescription  = "Extra attribute key of the object",
        Documentation.SampleValues      = {"c_customExtraAttribute"}
        ]),
    optional extAttributeValue9 as (type text meta [
        Documentation.FieldCaption      = "Extra attribute Value 10",
        Documentation.FieldDescription  = "Value of the extra attribute key 10",
        Documentation.SampleValues      = {""}
        ]),
    optional getStateHistory as (type logical meta [
        Documentation.FieldCaption      = "Get state history",
        Documentation.FieldDescription  = "Include state history in object data",
        Documentation.SampleValues      = {"true", "false"}
        ])
        )
    as table meta [];

FindUnderProjectByParameters = (
    address as text,
    projectCode as text,
    optional baseClass as text,
    optional typeCode as text,
    optional typeName as text,
    optional objCode as text,
    optional revision as text,
    optional isOldRevision as text,
    optional checkedOutByUserId as text,
    optional name as text,
    optional description as text,
    optional comment as text,

    optional author as text,
    optional authorId as text,
    optional creator as text,
    optional creatorId as text,
    optional owner as text,
    optional ownerId as text,

    optional publicity as text,
    optional publicityName as text,
    optional classification as text,
    optional classificationName as text,
    optional status as text,
    optional statusName as text,

    // Is there a better way to handle dates?
    //optional dated as text,
    //optional creationTime as text,
    //optional modifiedTime as text,

    // How many fields would a customer need from extAttributes for a single search?
    // Is there a better way of handling these in the UI?
    optional extAttributeKey0 as text,
    optional extAttributeValue0 as text,
    optional extAttributeKey1 as text,
    optional extAttributeValue1 as text,
    optional extAttributeKey2 as text,
    optional extAttributeValue2 as text,
    optional extAttributeKey3 as text,
    optional extAttributeValue3 as text,
    optional extAttributeKey4 as text,
    optional extAttributeValue4 as text,
    optional extAttributeKey5 as text,
    optional extAttributeValue5 as text,
    optional extAttributeKey6 as text,
    optional extAttributeValue6 as text,
    optional extAttributeKey7 as text,
    optional extAttributeValue7 as text,
    optional extAttributeKey8 as text,
    optional extAttributeValue8 as text,
    optional extAttributeKey9 as text,
    optional extAttributeValue9 as text,

    optional getStateHistory as logical
    ) =>
    let
        // Convert null -> ""
        searchJSON = ObjectsSearchParametersToJSON(
            {
                if (baseClass = null)           then {"",""} else {"baseClass",baseClass},
                if (typeCode = null)            then {"",""} else {"type",typeCode},
                if (typeName = null)            then {"",""} else {"typeName",typeName},
                if (objCode = null)             then {"",""} else {"objCode",objCode},
                if (revision = null)            then {"",""} else {"revision",revision},
                if (isOldRevision = null)       then {"",""} else {"isOldRevision",isOldRevision},
                if (checkedOutByUserId = null)  then {"",""} else {"checkedOutByUserId",checkedOutByUserId},
                if (name = null)                then {"",""} else {"name",name},
                if (description = null)         then {"",""} else {"description",description},
                if (comment = null)             then {"",""} else {"comment",comment},
                if (author = null)              then {"",""} else {"author",author},
                if (authorId = null)            then {"",""} else {"authorId",authorId},
                if (creator = null)             then {"",""} else {"creator",creator},
                if (creatorId = null)           then {"",""} else {"creatorId",creatorId},
                if (owner = null)               then {"",""} else {"owner",owner},
                if (ownerId = null)             then {"",""} else {"ownerId",ownerId},
                if (publicity = null)           then {"",""} else {"publicity",publicity},
                if (publicityName = null)       then {"",""} else {"publicityName",publicityName},
                if (classification = null)      then {"",""} else {"classification",classification},
                if (classificationName = null)  then {"",""} else {"classificationName",classificationName},
                if (status = null)              then {"",""} else {"status",status},
                if (statusName = null)          then {"",""} else {"statusName",statusName}
            },
            {
                if (extAttributeKey0 = null)    then {"",""} else {extAttributeKey0,extAttributeValue0},
                if (extAttributeKey1 = null)    then {"",""} else {extAttributeKey1,extAttributeValue1},
                if (extAttributeKey2 = null)    then {"",""} else {extAttributeKey2,extAttributeValue2},
                if (extAttributeKey3 = null)    then {"",""} else {extAttributeKey3,extAttributeValue3},
                if (extAttributeKey4 = null)    then {"",""} else {extAttributeKey4,extAttributeValue4},
                if (extAttributeKey5 = null)    then {"",""} else {extAttributeKey5,extAttributeValue5},
                if (extAttributeKey6 = null)    then {"",""} else {extAttributeKey6,extAttributeValue6},
                if (extAttributeKey7 = null)    then {"",""} else {extAttributeKey7,extAttributeValue7},
                if (extAttributeKey8 = null)    then {"",""} else {extAttributeKey8,extAttributeValue8},
                if (extAttributeKey9 = null)    then {"",""} else {extAttributeKey9,extAttributeValue9}
            }
            ),

        notEmptySearchJSON = if (searchJSON = "")
            then
                "{ }"
            else
                searchJSON,

        response = if (getStateHistory = null)
            then
                FindUnderProject(address, projectCode, notEmptySearchJSON)
            else
                FindUnderProject(
                    address,
                    projectCode,
                    notEmptySearchJSON,
                    {{"get-state-history", Logical.ToText(getStateHistory)}})
    in
        response;

FindUnderProject = (
    address as text,
    projectCode as text,
    jsonParameters as text,
    optional queryParameters as list) =>
    let
        table = GetTableThroughAuthorization(
            address,
            "projects/" & projectCode & "/find",
            jsonParameters,
            queryParameters)
    in
        table;

GetProjects = (address as text) as table =>
    let
        parameters =
            "{
                ""type"" : ""project""
            }",
    
        table = GetTableThroughAuthorization(
            address,
            "objects/find",
            parameters)
    in
        table;


GetProjectsNested = (address as text) as table =>
    let
        projectsTable = GetProjects(address),

        objCodes = Table.Column(projectsTable, "objCode"),
        names = Table.Column(projectsTable, "name"),
        data = List.Transform(objCodes, each GetObjectConnectionsNavTable(address, _)),

        itemKind = List.Transform(objCodes, each "Folder"),
        itemName = List.Transform(objCodes, each "Folder"),
        isLeaf = List.Transform(objCodes, each false),

        tableHeader = {"Name", "Key", "Data", "ItemKind", "ItemName", "IsLeaf"},
        table = Table.FromColumns({names, names, data, itemKind, itemName, isLeaf}, tableHeader),

        NavTable = Table.ToNavigationTable(table, {"Key"}, "Name", "Data", "ItemKind", "ItemName", "IsLeaf")
    in
        NavTable;

GetProjectsRecursive = (address as text) as table =>
    let
        projectsTable = GetProjects(address),

        objCodes = Table.Column(projectsTable, "objCode"),
        names = Table.Column(projectsTable, "name"),
        data = List.Transform(objCodes, each GetObjectConnectionsNavTableRecursive(address, _)),

        itemKind = List.Transform(objCodes, each "Folder"),
        itemName = List.Transform(objCodes, each "Folder"),
        isLeaf = List.Transform(objCodes, each false),

        tableHeader = {"Name", "Key", "Data", "ItemKind", "ItemName", "IsLeaf"},
        table = Table.FromColumns({names, names, data, itemKind, itemName, isLeaf}, tableHeader),

        NavTable = Table.ToNavigationTable(table, {"Key"}, "Name", "Data", "ItemKind", "ItemName", "IsLeaf")
    in
        NavTable;


/************************************
* Objects
************************************/

ObjectsGetSingleByTypeAndObjCode = (
  address as text,
  typeCode as text,
  objCode as text,
  optional getStateHistory as logical) as any =>
    let
        parameters =
            "{" &
                """type"" : """ & typeCode & """," &
                """objCode"" : """ & objCode & """
            }",

        getStateHistoryNotNull = if (getStateHistory = null)
          then
            false
          else
            getStateHistory,

        json = GetJsonThroughAuthorization(
            address,
            "objects/getSingle",
            parameters,
            {{"get-state-history", Logical.ToText(getStateHistoryNotNull)}})
    in
        json;

ObjectsGetSingleByParameters = (
    address as text,
    optional id as text,
    optional baseClass as text,
    optional typeCode as text,
    optional typeName as text,
    optional objCode as text,
    optional revision as text,
    optional revisionCount as text,
    optional isOldRevision as text,
    optional checkedOutByUserId as text,
    optional label as text,
    optional name as text,
    optional description as text,
    optional comment as text,

    optional author as text,
    optional authorId as text,
    optional creator as text,
    optional creatorId as text,
    optional owner as text,
    optional ownerId as text,

    optional publicity as text,
    optional publicityName as text,
    optional classification as text,
    optional classificationName as text,
    optional status as text,
    optional statusName as text,
    optional dictionaryId1 as text,

    // Is there a better way to handle dates?
    optional dated as text,
    optional creationTime as text,
    optional modifiedTime as text,

    // How many fields would a customer need from extAttributes for a single search?
    // Is there a better way of handling these in the UI?
    optional extAttributeKey0 as text,
    optional extAttributeValue0 as text,
    optional extAttributeKey1 as text,
    optional extAttributeValue1 as text,
    optional extAttributeKey2 as text,
    optional extAttributeValue2 as text,
    optional extAttributeKey3 as text,
    optional extAttributeValue3 as text,
    optional extAttributeKey4 as text,
    optional extAttributeValue4 as text,
    optional extAttributeKey5 as text,
    optional extAttributeValue5 as text,
    optional extAttributeKey6 as text,
    optional extAttributeValue6 as text,
    optional extAttributeKey7 as text,
    optional extAttributeValue7 as text,
    optional extAttributeKey8 as text,
    optional extAttributeValue8 as text,
    optional extAttributeKey9 as text,
    optional extAttributeValue9 as text
    ) as any =>
    let
        parameters =
            "{
                ""objCode"" : """ & objCode  & """
            }",

        response = GetJsonThroughAuthorization(
            address,
            "objects/getSingle",
            parameters)
    in
        response;

ObjectsGetSingle = (address as text, jsonParameters as text) as any =>
    let
        response = GetJsonThroughAuthorization(
            address,
            "objects/getSingle",
            jsonParameters)
    in
        response;

ObjectsGetStateHistory = (address as text, typeCode as text, objCode as text) =>
    let
        table = GetTableThroughAuthorization(
            address,
            "objects/status/"&typeCode&"/"&objCode&"/history"),

        expandStatus = Table.ExpandRecordColumn(
            table,
            "status",
            {"statusId", "statusName", "properties"},
            {"statusId", "statusName", "properties"}),

        result = if (Table.IsEmpty(table))
            then
                table
            else
                expandStatus
    in
        result;

TypeAndObjCodeToTableHandlerType = type function (
    typeCode as (type text meta [
        Documentation.FieldCaption      = "Type (code)",
        Documentation.FieldDescription  = "Object type",
        Documentation.SampleValues      = {"project"}
        ]),
    objCode as (type text meta [
        Documentation.FieldCaption      = "Object code",
        Documentation.FieldDescription  = "Object code",
        Documentation.SampleValues      = {"P000004"}
        ])
        
        )
    as table meta [];

TypeAndObjCodeToTableHandlerWithGetStateHistoryType = type function (
    typeCode as (type text meta [
        Documentation.FieldCaption      = "Type (code)",
        Documentation.FieldDescription  = "Object type",
        Documentation.SampleValues      = {"project"}
        ]),
    objCode as (type text meta [
        Documentation.FieldCaption      = "Object code",
        Documentation.FieldDescription  = "Object code",
        Documentation.SampleValues      = {"P000004"}
        ]),
    optional getStateHistory as (type logical meta [
        Documentation.FieldCaption      = "Get state history",
        Documentation.FieldDescription  = "Include state history in object data",
        Documentation.SampleValues      = {"true", "false"}
        ])
        )
    as table meta [];

ObjectsFindHandlerType = type function (
    jsonParameters as (type text meta [
        Documentation.FieldCaption      = "JSON Parameters",
        Documentation.FieldDescription  = "Search conditions",
        Documentation.SampleValues      = {"{ ""objCode"": ""P000004"" }"}
        ]),
    optional getStateHistory as (type logical meta [
        Documentation.FieldCaption      = "Get state history",
        Documentation.FieldDescription  = "Include state history in object data",
        Documentation.SampleValues      = {"true", "false"}
        ])
        )
    as table meta [];

ObjectsFind = (address as text, jsonParameters as text, optional queryParameters as list) =>
    let
        table = GetTableThroughAuthorization(
            address,
            "objects/find",
            jsonParameters,
            queryParameters)
    in
        table;

// Does not include start and end curly brackets
KeyValuePairsToJSON = (
    keyValuePairs as list) as text =>
      let
          jsonPartsWithEmpties = List.Transform(
              keyValuePairs,
              each
              if (_{0} <> "")
                  then
                      """" & _{0} & """: """ & _{1} & """"
                  else
                      ""
              ),
          jsonParts = List.RemoveItems(jsonPartsWithEmpties, {""}),

          // Combine the parts into a single-line JSON monstrosity
          jsonFinal = if (List.Count(jsonParts) = 0)
              then
                  ""
              else
                  Text.Combine(jsonParts, ", ")
      in
          jsonFinal;


ObjectsSearchParametersToJSON = (
    keyValuePairs as list,
    extAttributesKeyValuePairs as list) =>
      let
          basicFields = KeyValuePairsToJSON(keyValuePairs),
          extAttributes = KeyValuePairsToJSON(extAttributesKeyValuePairs),

          jsonParts =
              if (extAttributes <> "")
                then basicFields & ", ""extAttributes"": { " & extAttributes & " }"
                else basicFields
              ,

          // Combine the parts into a single-line JSON monstrosity
          jsonFinal = if (jsonParts <> "")
              then
                  "{ " & jsonParts & " }"
              else
                  ""
      in
          jsonFinal;

ObjectsFindByParametersHandlerType = type function (
    optional baseClass as (type text meta [
        Documentation.FieldCaption      = "Base class",
        Documentation.FieldDescription  = "Base class of the object",
        Documentation.SampleValues      = {"doc"}
        ]),
    optional typeCode as (type text meta [
        Documentation.FieldCaption      = "Type (code)",
        Documentation.FieldDescription  = "Type of the object (type code)",
        Documentation.SampleValues      = {"model"}
        ]),
//     optional typeName as (type text meta [
//         Documentation.FieldCaption      = "Type",
//         Documentation.FieldDescription  = "Type of the object (type name)",
//         Documentation.SampleValues      = {"Model (Vertex)"}
//         ]),
    optional objCode as (type text meta [
        Documentation.FieldCaption      = "Object code",
        Documentation.FieldDescription  = "Object code of the object",
        Documentation.SampleValues      = {"P000004"}
        ]),
    optional revision as (type text meta [
        Documentation.FieldCaption      = "Revision",
        Documentation.FieldDescription  = "Revision of the object",
        Documentation.SampleValues      = {"0"}
        ]),
//     optional isOldRevision as (type text meta [
//         Documentation.FieldCaption      = "Is old revision",
//         Documentation.FieldDescription  = "False if this is the newest revision of the object",
//         Documentation.SampleValues      = {"false"}
//         ]),
//     optional checkedOutByUserId as (type text meta [
//         Documentation.FieldCaption      = "Checked out by user ID",
//         Documentation.FieldDescription  = "The object is checked out by a specific user",
//         Documentation.SampleValues      = {""}
//         ]),
    optional name as (type text meta [
        Documentation.FieldCaption      = "Name",
        Documentation.FieldDescription  = "Name of the object",
        Documentation.SampleValues      = {""}
        ]),
    optional description as (type text meta [
        Documentation.FieldCaption      = "Description",
        Documentation.FieldDescription  = "Description of the object",
        Documentation.SampleValues      = {""}
        ]),
//     optional comment as (type text meta [
//         Documentation.FieldCaption      = "Comment",
//         Documentation.FieldDescription  = "Comment of the object",
//         Documentation.SampleValues      = {""}
//         ]),
//     optional author as (type text meta [
//         Documentation.FieldCaption      = "Author",
//         Documentation.FieldDescription  = "Author of the object",
//         Documentation.SampleValues      = {""}
//         ]),
    optional authorId as (type text meta [
        Documentation.FieldCaption      = "Author ID",
        Documentation.FieldDescription  = "ID of the author of the object",
        Documentation.SampleValues      = {""}
        ]),
//     optional creator as (type text meta [
//         Documentation.FieldCaption      = "Creator",
//         Documentation.FieldDescription  = "Creator of the object",
//         Documentation.SampleValues      = {""}
//         ]),
    optional creatorId as (type text meta [
        Documentation.FieldCaption      = "Creator ID",
        Documentation.FieldDescription  = "ID of the creator of the object",
        Documentation.SampleValues      = {""}
        ]),
//     optional owner as (type text meta [
//         Documentation.FieldCaption      = "Owner",
//         Documentation.FieldDescription  = "Owner of the object",
//         Documentation.SampleValues      = {""}
//         ]),
//     optional ownerId as (type text meta [
//         Documentation.FieldCaption      = "Owner ID",
//         Documentation.FieldDescription  = "ID of the owner of the object",
//         Documentation.SampleValues      = {""}
//         ]),
//     optional publicity as (type text meta [
//         Documentation.FieldCaption      = "Publicity (code)",
//         Documentation.FieldDescription  = "Publicity of the object (publicity code)",
//         Documentation.SampleValues      = {""}
//         ]),
//     optional publicityName as (type text meta [
//         Documentation.FieldCaption      = "Publicity",
//         Documentation.FieldDescription  = "Publicity of the object (publicity name)",
//         Documentation.SampleValues      = {""}
//         ]),
    optional classification as (type text meta [
        Documentation.FieldCaption      = "Classification (code)",
        Documentation.FieldDescription  = "Classification of the object (classification code)",
        Documentation.SampleValues      = {""}
        ]),
//     optional classificationName as (type text meta [
//         Documentation.FieldCaption      = "Classification",
//         Documentation.FieldDescription  = "Classification of the object (classification code)",
//         Documentation.SampleValues      = {""}
//         ]),
    optional status as (type text meta [
        Documentation.FieldCaption      = "Status (code)",
        Documentation.FieldDescription  = "Status of the object",
        Documentation.SampleValues      = {"DRAFT"}
        ]),
//     optional statusName as (type text meta [
//         Documentation.FieldCaption      = "Status",
//         Documentation.FieldDescription  = "Status of the object (status name)",
//         Documentation.SampleValues      = {"Draft"}
//         ]),
    // Notice off-by-one (programmers start from 0, people start from 1)
    optional extAttributeKey0 as (type text meta [
        Documentation.FieldCaption      = "Extra attribute Key 1",
        Documentation.FieldDescription  = "Extra attribute key of the object",
        Documentation.SampleValues      = {"c_customExtraAttribute"}
        ]),
    optional extAttributeValue0 as (type text meta [
        Documentation.FieldCaption      = "Extra attribute Value 1",
        Documentation.FieldDescription  = "Value of the extra attribute key 1",
        Documentation.SampleValues      = {""}
        ]),
        optional extAttributeKey1 as (type text meta [
        Documentation.FieldCaption      = "Extra attribute Key 2",
        Documentation.FieldDescription  = "Extra attribute key of the object",
        Documentation.SampleValues      = {"c_customExtraAttribute"}
        ]),
    optional extAttributeValue1 as (type text meta [
        Documentation.FieldCaption      = "Extra attribute Value 2",
        Documentation.FieldDescription  = "Value of the extra attribute key 2",
        Documentation.SampleValues      = {""}
        ]),
        optional extAttributeKey2 as (type text meta [
        Documentation.FieldCaption      = "Extra attribute Key 3",
        Documentation.FieldDescription  = "Extra attribute key of the object",
        Documentation.SampleValues      = {"c_customExtraAttribute"}
        ]),
    optional extAttributeValue2 as (type text meta [
        Documentation.FieldCaption      = "Extra attribute Value 3",
        Documentation.FieldDescription  = "Value of the extra attribute key 3",
        Documentation.SampleValues      = {""}
        ]),
        optional extAttributeKey3 as (type text meta [
        Documentation.FieldCaption      = "Extra attribute Key 4",
        Documentation.FieldDescription  = "Extra attribute key of the object",
        Documentation.SampleValues      = {"c_customExtraAttribute"}
        ]),
    optional extAttributeValue3 as (type text meta [
        Documentation.FieldCaption      = "Extra attribute Value 4",
        Documentation.FieldDescription  = "Value of the extra attribute key 4",
        Documentation.SampleValues      = {""}
        ]),
        optional extAttributeKey4 as (type text meta [
        Documentation.FieldCaption      = "Extra attribute Key 5",
        Documentation.FieldDescription  = "Extra attribute key of the object",
        Documentation.SampleValues      = {"c_customExtraAttribute"}
        ]),
    optional extAttributeValue4 as (type text meta [
        Documentation.FieldCaption      = "Extra attribute Value 5",
        Documentation.FieldDescription  = "Value of the extra attribute key 5",
        Documentation.SampleValues      = {""}
        ]),
        optional extAttributeKey5 as (type text meta [
        Documentation.FieldCaption      = "Extra attribute Key 6",
        Documentation.FieldDescription  = "Extra attribute key of the object",
        Documentation.SampleValues      = {"c_customExtraAttribute"}
        ]),
    optional extAttributeValue5 as (type text meta [
        Documentation.FieldCaption      = "Extra attribute Value 6",
        Documentation.FieldDescription  = "Value of the extra attribute key 6",
        Documentation.SampleValues      = {""}
        ]),
        optional extAttributeKey6 as (type text meta [
        Documentation.FieldCaption      = "Extra attribute Key 7",
        Documentation.FieldDescription  = "Extra attribute key of the object",
        Documentation.SampleValues      = {"c_customExtraAttribute"}
        ]),
    optional extAttributeValue6 as (type text meta [
        Documentation.FieldCaption      = "Extra attribute Value 7",
        Documentation.FieldDescription  = "Value of the extra attribute key 7",
        Documentation.SampleValues      = {""}
        ]),
        optional extAttributeKey7 as (type text meta [
        Documentation.FieldCaption      = "Extra attribute Key 8",
        Documentation.FieldDescription  = "Extra attribute key of the object",
        Documentation.SampleValues      = {"c_customExtraAttribute"}
        ]),
    optional extAttributeValue7 as (type text meta [
        Documentation.FieldCaption      = "Extra attribute Value 8",
        Documentation.FieldDescription  = "Value of the extra attribute key 8",
        Documentation.SampleValues      = {""}
        ]),
        optional extAttributeKey8 as (type text meta [
        Documentation.FieldCaption      = "Extra attribute Key 9",
        Documentation.FieldDescription  = "Extra attribute key of the object",
        Documentation.SampleValues      = {"c_customExtraAttribute"}
        ]),
    optional extAttributeValue8 as (type text meta [
        Documentation.FieldCaption      = "Extra attribute Value 9",
        Documentation.FieldDescription  = "Value of the extra attribute key 9",
        Documentation.SampleValues      = {""}
        ]),
        optional extAttributeKey9 as (type text meta [
        Documentation.FieldCaption      = "Extra attribute Key 10",
        Documentation.FieldDescription  = "Extra attribute key of the object",
        Documentation.SampleValues      = {"c_customExtraAttribute"}
        ]),
    optional extAttributeValue9 as (type text meta [
        Documentation.FieldCaption      = "Extra attribute Value 10",
        Documentation.FieldDescription  = "Value of the extra attribute key 10",
        Documentation.SampleValues      = {""}
        ]),
    optional getStateHistory as (type logical meta [
        Documentation.FieldCaption      = "Get state history",
        Documentation.FieldDescription  = "Include state history in object data",
        Documentation.SampleValues      = {"true", "false"}
        ])
        )
    as table meta [];

// So many parameters...
ObjectsFindByParameters = (
    address as text,
    optional baseClass as text,
    optional typeCode as text,
    //optional typeName as text,
    optional objCode as text,
    optional revision as text,
    //optional isOldRevision as text,
    //optional checkedOutByUserId as text,
    optional name as text,
    optional description as text,
    //optional comment as text,

    //optional author as text,
    optional authorId as text,
    //optional creator as text,
    optional creatorId as text,
    //optional owner as text,
    //optional ownerId as text,

    //optional publicity as text,
    //optional publicityName as text,
    optional classification as text,
    //optional classificationName as text,
    optional status as text,
    //optional statusName as text,

    // Is there a better way to handle dates?
    //optional dated as text,
    //optional creationTime as text,
    //optional modifiedTime as text,

    // How many fields would a customer need from extAttributes for a single search?
    // Is there a better way of handling these in the UI?
    optional extAttributeKey0 as text,
    optional extAttributeValue0 as text,
    optional extAttributeKey1 as text,
    optional extAttributeValue1 as text,
    optional extAttributeKey2 as text,
    optional extAttributeValue2 as text,
    optional extAttributeKey3 as text,
    optional extAttributeValue3 as text,
    optional extAttributeKey4 as text,
    optional extAttributeValue4 as text,
    optional extAttributeKey5 as text,
    optional extAttributeValue5 as text,
    optional extAttributeKey6 as text,
    optional extAttributeValue6 as text,
    optional extAttributeKey7 as text,
    optional extAttributeValue7 as text,
    optional extAttributeKey8 as text,
    optional extAttributeValue8 as text,
    optional extAttributeKey9 as text,
    optional extAttributeValue9 as text,

    optional getStateHistory as logical
    ) =>
    let
        // Convert null -> ""
        searchJSON = ObjectsSearchParametersToJSON(
            {
                if (baseClass = null)           then {"",""} else {"baseClass",baseClass},
                if (typeCode = null)            then {"",""} else {"type",typeCode},
                //if (typeName = null)            then {"",""} else {"typeName",typeName},
                if (objCode = null)             then {"",""} else {"objCode",objCode},
                if (revision = null)            then {"",""} else {"revision",revision},
                //if (isOldRevision = null)       then {"",""} else {"isOldRevision",isOldRevision},
                //if (checkedOutByUserId = null)  then {"",""} else {"checkedOutByUserId",checkedOutByUserId},
                if (name = null)                then {"",""} else {"name",name},
                if (description = null)         then {"",""} else {"description",description},
                //if (comment = null)             then {"",""} else {"comment",comment},
                //if (author = null)              then {"",""} else {"author",author},
                if (authorId = null)            then {"",""} else {"authorId",authorId},
                //if (creator = null)             then {"",""} else {"creator",creator},
                if (creatorId = null)           then {"",""} else {"creatorId",creatorId},
                //if (owner = null)               then {"",""} else {"owner",owner},
                //if (ownerId = null)             then {"",""} else {"ownerId",ownerId},
                //if (publicity = null)           then {"",""} else {"publicity",publicity},
                //if (publicityName = null)       then {"",""} else {"publicityName",publicityName},
                if (classification = null)      then {"",""} else {"classification",classification},
                //if (classificationName = null)  then {"",""} else {"classificationName",classificationName},
                if (status = null)              then {"",""} else {"status",status}//,
                //if (statusName = null)          then {"",""} else {"statusName",statusName}
            },
            {
                if (extAttributeKey0 = null)    then {"",""} else {extAttributeKey0,extAttributeValue0},
                if (extAttributeKey1 = null)    then {"",""} else {extAttributeKey1,extAttributeValue1},
                if (extAttributeKey2 = null)    then {"",""} else {extAttributeKey2,extAttributeValue2},
                if (extAttributeKey3 = null)    then {"",""} else {extAttributeKey3,extAttributeValue3},
                if (extAttributeKey4 = null)    then {"",""} else {extAttributeKey4,extAttributeValue4},
                if (extAttributeKey5 = null)    then {"",""} else {extAttributeKey5,extAttributeValue5},
                if (extAttributeKey6 = null)    then {"",""} else {extAttributeKey6,extAttributeValue6},
                if (extAttributeKey7 = null)    then {"",""} else {extAttributeKey7,extAttributeValue7},
                if (extAttributeKey8 = null)    then {"",""} else {extAttributeKey8,extAttributeValue8},
                if (extAttributeKey9 = null)    then {"",""} else {extAttributeKey9,extAttributeValue9}
            }
            ),

        response = if (searchJSON = "")
            then
                ""
            else
                if (getStateHistory = null)
                    then ObjectsFind(address, searchJSON)
                    else ObjectsFind(address, searchJSON, {{"get-state-history",Logical.ToText(getStateHistory)}})
    in
        response;

ObjectsConnectionsHandlerType = type function (
    objType as (type text meta [
        Documentation.FieldCaption      = "Object type",
        Documentation.FieldDescription  = "Type of the object",
        Documentation.SampleValues      = {"project"}
        ]),
    objCode as (type text meta [
        Documentation.FieldCaption      = "Object code",
        Documentation.FieldDescription  = "Object code of the object",
        Documentation.SampleValues      = {"P000004"}
        ]),
    optional getStateHistory as (type logical meta [
        Documentation.FieldCaption      = "Get state history",
        Documentation.FieldDescription  = "Include state history in object data",
        Documentation.SampleValues      = {"true", "false"}
        ])
    )
    as table meta [];

ObjectsFindInConnectionsHandlerType = type function (
    objType as (type text meta [
        Documentation.FieldCaption      = "Parent object type",
        Documentation.FieldDescription  = "Type of the parent object",
        Documentation.SampleValues      = {"project"}
        ]),
    objCode as (type text meta [
        Documentation.FieldCaption      = "Parent object code",
        Documentation.FieldDescription  = "Object code of the parent object",
        Documentation.SampleValues      = {"P000004"}
        ]),
    optional filterBaseClass as (type text meta [
        Documentation.FieldCaption      = "Filter base class",
        Documentation.FieldDescription  = "Filter base class of the objects",
        Documentation.SampleValues      = {"doc"}
        ]),
    optional filterTypeCode as (type text meta [
        Documentation.FieldCaption      = "Filter type (code)",
        Documentation.FieldDescription  = "Filter type of the objects (type code)",
        Documentation.SampleValues      = {"model"}
        ]),
    optional filterTypeName as (type text meta [
        Documentation.FieldCaption      = "Filter type",
        Documentation.FieldDescription  = "Filter type of the objects (type name)",
        Documentation.SampleValues      = {"Model (Vertex)"}
        ]),
    optional filterObjCode as (type text meta [
        Documentation.FieldCaption      = "Filter object code",
        Documentation.FieldDescription  = "Filter object code of the objects",
        Documentation.SampleValues      = {"P000004"}
        ]),
    optional filterRevision as (type text meta [
        Documentation.FieldCaption      = "Filter revision",
        Documentation.FieldDescription  = "Filter revision of the objects",
        Documentation.SampleValues      = {"0"}
        ]),
    optional filterIsOldRevision as (type text meta [
        Documentation.FieldCaption      = "Filter is old revision",
        Documentation.FieldDescription  = "False if these are the newest revisions of the objects",
        Documentation.SampleValues      = {"false"}
        ]),
    optional filterCheckedOutByUserId as (type text meta [
        Documentation.FieldCaption      = "Filter checked out by user ID",
        Documentation.FieldDescription  = "Filter objects that are checked out by a specific user",
        Documentation.SampleValues      = {""}
        ]),
    optional filterName as (type text meta [
        Documentation.FieldCaption      = "Filter name",
        Documentation.FieldDescription  = "Filter name of the objects",
        Documentation.SampleValues      = {""}
        ]),
    optional filterDescription as (type text meta [
        Documentation.FieldCaption      = "Filter description",
        Documentation.FieldDescription  = "Filter description of the objects",
        Documentation.SampleValues      = {""}
        ]),
    optional filterComment as (type text meta [
        Documentation.FieldCaption      = "Filter comment",
        Documentation.FieldDescription  = "Filter comment of the objects",
        Documentation.SampleValues      = {""}
        ]),
    optional filterAuthor as (type text meta [
        Documentation.FieldCaption      = "Filter author",
        Documentation.FieldDescription  = "Filter author of the objects",
        Documentation.SampleValues      = {""}
        ]),
    optional filterAuthorId as (type text meta [
        Documentation.FieldCaption      = "Filter author ID",
        Documentation.FieldDescription  = "Filter ID of the author of the objects",
        Documentation.SampleValues      = {""}
        ]),
    optional filterCreator as (type text meta [
        Documentation.FieldCaption      = "Filter creator",
        Documentation.FieldDescription  = "Filter creator of the objects",
        Documentation.SampleValues      = {""}
        ]),
    optional filterCreatorId as (type text meta [
        Documentation.FieldCaption      = "Filter creator ID",
        Documentation.FieldDescription  = "Filter ID of the creator of the objects",
        Documentation.SampleValues      = {""}
        ]),
    optional filterOwner as (type text meta [
        Documentation.FieldCaption      = "Filter owner",
        Documentation.FieldDescription  = "Filter owner of the objects",
        Documentation.SampleValues      = {""}
        ]),
    optional filterOwnerId as (type text meta [
        Documentation.FieldCaption      = "Filter owner ID",
        Documentation.FieldDescription  = "Filter ID of the owner of the objects",
        Documentation.SampleValues      = {""}
        ]),
    optional filterPublicity as (type text meta [
        Documentation.FieldCaption      = "Filter publicity (code)",
        Documentation.FieldDescription  = "Filter publicity of the objects (publicity code)",
        Documentation.SampleValues      = {""}
        ]),
    optional filterPublicityName as (type text meta [
        Documentation.FieldCaption      = "Filter publicity",
        Documentation.FieldDescription  = "Filter publicity of the objects (publicity name)",
        Documentation.SampleValues      = {""}
        ]),
    optional filterClassification as (type text meta [
        Documentation.FieldCaption      = "Filter classification (code)",
        Documentation.FieldDescription  = "Filter classification of the objects (classification code)",
        Documentation.SampleValues      = {""}
        ]),
    optional filterClassificationName as (type text meta [
        Documentation.FieldCaption      = "Filter classification",
        Documentation.FieldDescription  = "Filter classification of the objects (classification code)",
        Documentation.SampleValues      = {""}
        ]),
    optional filterStatus as (type text meta [
        Documentation.FieldCaption      = "Filter status (code)",
        Documentation.FieldDescription  = "Filter status of the objects",
        Documentation.SampleValues      = {"DRAFT"}
        ]),
    optional filterStatusName as (type text meta [
        Documentation.FieldCaption      = "Filter status",
        Documentation.FieldDescription  = "Filter status of the objects (status name)",
        Documentation.SampleValues      = {"Draft"}
        ]),
    //optional filterCreationTime as (type text meta [
    //    Documentation.FieldCaption      = "Filter creation time",
    //    Documentation.FieldDescription  = "Filter creation time of the objects",
    //    Documentation.SampleValues      = {""}
    //    ]),
    //optional filterModifiedTime as (type text meta [
    //    Documentation.FieldCaption      = "Filter modified time",
    //    Documentation.FieldDescription  = "Filter modified of the objects",
    //    Documentation.SampleValues      = {""}
    //    ]),
    // Notice off-by-one (programmers start from 0, people start from 1)
    optional filterExtAttributeKey0 as (type text meta [
        Documentation.FieldCaption      = "Filter extra attribute Key 1",
        Documentation.FieldDescription  = "Filter extra attribute key of the objects",
        Documentation.SampleValues      = {"c_customExtraAttribute"}
        ]),
    optional filterExtAttributeValue0 as (type text meta [
        Documentation.FieldCaption      = "Filter extra attribute Value 1",
        Documentation.FieldDescription  = "Filter value of the extra attribute key 1",
        Documentation.SampleValues      = {""}
        ]),
        optional filterExtAttributeKey1 as (type text meta [
        Documentation.FieldCaption      = "Filter extra attribute Key 2",
        Documentation.FieldDescription  = "Filter extra attribute key of the objects",
        Documentation.SampleValues      = {"c_customExtraAttribute"}
        ]),
    optional filterExtAttributeValue1 as (type text meta [
        Documentation.FieldCaption      = "Filter extra attribute Value 2",
        Documentation.FieldDescription  = "Filter value of the extra attribute key 2",
        Documentation.SampleValues      = {""}
        ]),
        optional filterExtAttributeKey2 as (type text meta [
        Documentation.FieldCaption      = "Filter extra attribute Key 3",
        Documentation.FieldDescription  = "Filter extra attribute key of the objects",
        Documentation.SampleValues      = {"c_customExtraAttribute"}
        ]),
    optional filterExtAttributeValue2 as (type text meta [
        Documentation.FieldCaption      = "Filter extra attribute Value 3",
        Documentation.FieldDescription  = "Filter value of the extra attribute key 3",
        Documentation.SampleValues      = {""}
        ]),
        optional filterExtAttributeKey3 as (type text meta [
        Documentation.FieldCaption      = "Filter extra attribute Key 4",
        Documentation.FieldDescription  = "Filter extra attribute key of the objects",
        Documentation.SampleValues      = {"c_customExtraAttribute"}
        ]),
    optional filterExtAttributeValue3 as (type text meta [
        Documentation.FieldCaption      = "Filter extra attribute Value 4",
        Documentation.FieldDescription  = "Filter value of the extra attribute key 4",
        Documentation.SampleValues      = {""}
        ]),
        optional filterExtAttributeKey4 as (type text meta [
        Documentation.FieldCaption      = "Filter extra attribute Key 5",
        Documentation.FieldDescription  = "Filter extra attribute key of the objects",
        Documentation.SampleValues      = {"c_customExtraAttribute"}
        ]),
    optional filterExtAttributeValue4 as (type text meta [
        Documentation.FieldCaption      = "Filter extra attribute Value 5",
        Documentation.FieldDescription  = "Filter value of the extra attribute key 5",
        Documentation.SampleValues      = {""}
        ]),
        optional filterExtAttributeKey5 as (type text meta [
        Documentation.FieldCaption      = "Filter extra attribute Key 6",
        Documentation.FieldDescription  = "Filter extra attribute key of the objects",
        Documentation.SampleValues      = {"c_customExtraAttribute"}
        ]),
    optional filterExtAttributeValue5 as (type text meta [
        Documentation.FieldCaption      = "Filter extra attribute Value 6",
        Documentation.FieldDescription  = "Filter value of the extra attribute key 6",
        Documentation.SampleValues      = {""}
        ]),
        optional filterExtAttributeKey6 as (type text meta [
        Documentation.FieldCaption      = "Filter extra attribute Key 7",
        Documentation.FieldDescription  = "Filter extra attribute key of the objects",
        Documentation.SampleValues      = {"c_customExtraAttribute"}
        ]),
    optional filterExtAttributeValue6 as (type text meta [
        Documentation.FieldCaption      = "Filter extra attribute Value 7",
        Documentation.FieldDescription  = "Filter value of the extra attribute key 7",
        Documentation.SampleValues      = {""}
        ]),
        optional filterExtAttributeKey7 as (type text meta [
        Documentation.FieldCaption      = "Filter extra attribute Key 8",
        Documentation.FieldDescription  = "Filter extra attribute key of the objects",
        Documentation.SampleValues      = {"c_customExtraAttribute"}
        ]),
    optional filterExtAttributeValue7 as (type text meta [
        Documentation.FieldCaption      = "Filter extra attribute Value 8",
        Documentation.FieldDescription  = "Filter value of the extra attribute key 8",
        Documentation.SampleValues      = {""}
        ]),
        optional filterExtAttributeKey8 as (type text meta [
        Documentation.FieldCaption      = "Filter extra attribute Key 9",
        Documentation.FieldDescription  = "Filter extra attribute key of the objects",
        Documentation.SampleValues      = {"c_customExtraAttribute"}
        ]),
    optional filterExtAttributeValue8 as (type text meta [
        Documentation.FieldCaption      = "Filter extra attribute Value 9",
        Documentation.FieldDescription  = "Filter value of the extra attribute key 9",
        Documentation.SampleValues      = {""}
        ]),
        optional filterExtAttributeKey9 as (type text meta [
        Documentation.FieldCaption      = "Filter extra attribute Key 10",
        Documentation.FieldDescription  = "Filter extra attribute key of the objects",
        Documentation.SampleValues      = {"c_customExtraAttribute"}
        ]),
    optional filterExtAttributeValue9 as (type text meta [
        Documentation.FieldCaption      = "Filter extra attribute Value 10",
        Documentation.FieldDescription  = "Filter value of the extra attribute key 10",
        Documentation.SampleValues      = {""}
        ]),
    optional getStateHistory as (type logical meta [
        Documentation.FieldCaption      = "Get state history",
        Documentation.FieldDescription  = "Include state history in object data",
        Documentation.SampleValues      = {"true", "false"}
        ])
    )
    as table meta [];

GetObjectConnectionsDirect = (
  address as text,
  objType as text,
  objCode as text,
  optional getStateHistory as logical
  ) as table =>
    let
        getStateHistoryNotNull = if (getStateHistory = null)
          then
            false
          else
            getStateHistory,

        table = GetTableThroughAuthorization(
            address,
            "objects/" & objType & "/" & objCode & "/connections",
            null,
            {{"get-state-history", Logical.ToText(getStateHistoryNotNull)}})
    in
        table;

ObjectsFindInConnections = (
    address as text,
    objType as text,
    objCode as text,
    optional filterBaseClass as text,
    optional filterTypeCode as text,
    optional filterTypeName as text,
    optional filterObjCode as text,
    optional filterRevision as text,
    optional filterIsOldRevision as text,
    optional filterCheckedOutByUserId as text,
    optional filterName as text,
    optional filterDescription as text,
    optional filterComment as text,
    optional filterAuthor as text,
    optional filterAuthorId as text,
    optional filterCreator as text,
    optional filterCreatorId as text,
    optional filterOwner as text,
    optional filterOwnerId as text,
    optional filterPublicity as text,
    optional filterPublicityName as text,
    optional filterClassification as text,
    optional filterClassificationName as text,
    optional filterStatus as text,
    optional filterStatusName as text,
    //optional filterCreationTime as text,
    //optional filterModifiedTime as text,
    optional filterExtAttributeKey0 as text,
    optional filterExtAttributeValue0 as text,
    optional filterExtAttributeKey1 as text,
    optional filterExtAttributeValue1 as text,
    optional filterExtAttributeKey2 as text,
    optional filterExtAttributeValue2 as text,
    optional filterExtAttributeKey3 as text,
    optional filterExtAttributeValue3 as text,
    optional filterExtAttributeKey4 as text,
    optional filterExtAttributeValue4 as text,
    optional filterExtAttributeKey5 as text,
    optional filterExtAttributeValue5 as text,
    optional filterExtAttributeKey6 as text,
    optional filterExtAttributeValue6 as text,
    optional filterExtAttributeKey7 as text,
    optional filterExtAttributeValue7 as text,
    optional filterExtAttributeKey8 as text,
    optional filterExtAttributeValue8 as text,
    optional filterExtAttributeKey9 as text,
    optional filterExtAttributeValue9 as text,
    optional getStateHistory as logical
    ) as table =>
    let
        
        searchJSON = ObjectsSearchParametersToJSON(
            {
                if (filterBaseClass = null)           then {"",""} else {"baseClass",filterBaseClass},
                if (filterTypeCode = null)            then {"",""} else {"type",filterTypeCode},
                if (filterTypeName = null)            then {"",""} else {"typeName",filterTypeName},
                if (filterObjCode = null)             then {"",""} else {"objCode",filterObjCode},
                if (filterRevision = null)            then {"",""} else {"revision",filterRevision},
                if (filterIsOldRevision = null)       then {"",""} else {"isOldRevision",filterIsOldRevision},
                if (filterCheckedOutByUserId = null)  then {"",""} else {"checkedOutByUserId",filterCheckedOutByUserId},
                if (filterName = null)                then {"",""} else {"name",filterName},
                if (filterDescription = null)         then {"",""} else {"description",filterDescription},
                if (filterComment = null)             then {"",""} else {"comment",filterComment},
                if (filterAuthor = null)              then {"",""} else {"author",filterAuthor},
                if (filterAuthorId = null)            then {"",""} else {"authorId",filterAuthorId},
                if (filterCreator = null)             then {"",""} else {"creator",filterCreator},
                if (filterCreatorId = null)           then {"",""} else {"creatorId",filterCreatorId},
                if (filterOwner = null)               then {"",""} else {"owner",filterOwner},
                if (filterOwnerId = null)             then {"",""} else {"ownerId",filterOwnerId},
                if (filterPublicity = null)           then {"",""} else {"publicity",filterPublicity},
                if (filterPublicityName = null)       then {"",""} else {"publicityName",filterPublicityName},
                if (filterClassification = null)      then {"",""} else {"classification",filterClassification},
                if (filterClassificationName = null)  then {"",""} else {"classificationName",filterClassificationName},
                if (filterStatus = null)              then {"",""} else {"status",filterStatus},
                if (filterStatusName = null)          then {"",""} else {"statusName",filterStatusName} 
            },
            {
                if (filterExtAttributeKey0 = null)    then {"",""} else {filterExtAttributeKey0,filterExtAttributeValue0},
                if (filterExtAttributeKey1 = null)    then {"",""} else {filterExtAttributeKey1,filterExtAttributeValue1},
                if (filterExtAttributeKey2 = null)    then {"",""} else {filterExtAttributeKey2,filterExtAttributeValue2},
                if (filterExtAttributeKey3 = null)    then {"",""} else {filterExtAttributeKey3,filterExtAttributeValue3},
                if (filterExtAttributeKey4 = null)    then {"",""} else {filterExtAttributeKey4,filterExtAttributeValue4},
                if (filterExtAttributeKey5 = null)    then {"",""} else {filterExtAttributeKey5,filterExtAttributeValue5},
                if (filterExtAttributeKey6 = null)    then {"",""} else {filterExtAttributeKey6,filterExtAttributeValue6},
                if (filterExtAttributeKey7 = null)    then {"",""} else {filterExtAttributeKey7,filterExtAttributeValue7},
                if (filterExtAttributeKey8 = null)    then {"",""} else {filterExtAttributeKey8,filterExtAttributeValue8},
                if (filterExtAttributeKey9 = null)    then {"",""} else {filterExtAttributeKey9,filterExtAttributeValue9}
            }
            ),

        getStateHistoryNotNull = if (getStateHistory = null)
          then
            false
          else
            getStateHistory,

        table = GetTableThroughAuthorization(
            address,
            "objects/" & objType & "/" & objCode & "/connections",
            searchJSON,
            {{"get-state-history",Logical.ToText(getStateHistoryNotNull)}}
            )
    in
        table;

GetObjectConnectionsNavTable = (address as text, objCode as text) as table =>
    let
        parameters =
            "{
                ""objCode"" : """ & objCode  & """
            }",

        response = Json.Document(
            GetRequestThroughAuthorization(address,
            "objects/getSingle",
            parameters)),

        // First the project itself
        projectLabel    = {"_" & response[label]}, // Attempts to be the first in the alphabetical sorting
        projectKey      = {response[label]},
        projectData     = {response},
        projectItemKind = {"Cube"},
        projectItemName = {"Cube"},
        projectIsLeaf   = {true},

        // Then the connections of the project
        connectionsList = response[connections],
        connectionsTable = Table.FromRecords(connectionsList),

        labels = Table.Column(connectionsTable, "label"),

        types = Table.Column(connectionsTable, "type"),
        objCodes = Table.Column(connectionsTable, "objCode"),
        revisions = Table.Column(connectionsTable, "revision"),

        keysList = List.Zip({types, objCodes, revisions}),

        keys = List.Transform(keysList, each Text.Combine(_)),

        data = connectionsList,

        itemKind = List.Transform(labels, each "Table"),
        itemName = List.Transform(labels, each "Table"),
        isLeaf = List.Transform(labels, each true),

        allLabels   = List.Combine({projectLabel,    labels}),
        allKeys     = List.Combine({projectKey,      keys}),
        allData     = List.Combine({projectData,     data}),
        allItemKind = List.Combine({projectItemKind, itemKind}),
        allItemName = List.Combine({projectItemName, itemName}),
        allIsLeaf   = List.Combine({projectIsLeaf,   isLeaf}),

        tableHeader = {"Name", "Key", "Data", "ItemKind", "ItemName", "IsLeaf"},
        table = Table.FromColumns({allLabels, allKeys, allData, allItemKind, allItemName, allIsLeaf}, tableHeader),

        NavTable = Table.ToNavigationTable(table, {"Key"}, "Name", "Data", "ItemKind", "ItemName", "IsLeaf")
    in
        NavTable;

GetObjectConnectionsNavTableRecursive = (address as text, objCode as text) as table =>
    let
        parameters =
            "{
                ""objCode"" : """ & objCode  & """
            }",

        response = Json.Document(
            GetRequestThroughAuthorization(address,
            "objects/getSingle",
            parameters)),

        connectionsList = response[connections],
        connectionsTable = Table.FromRecords(connectionsList),

        // First the object itself
        objectLabel    = response[label],
        objectKey      = response[label],
        objectData     = response,
        objectItemKind = "View",
        objectItemName = "View",
        objectIsLeaf   = true,

        // Then just the first level of connections
        firstLevelConnectionsLabel    = "Connections",
        firstLevelConnectionsKey      = "connectionsFlat",
        firstLevelConnectionsData     = connectionsTable,
        firstLevelConnectionsItemKind = "Table",
        firstLevelConnectionsItemName = "Table",
        firstLevelConnectionsIsLeaf   = true,

        // Then the connections of the project recursively
        

        labels = Table.Column(connectionsTable, "label"),

        types = Table.Column(connectionsTable, "type"),
        objCodes = Table.Column(connectionsTable, "objCode"),
        revisions = Table.Column(connectionsTable, "revision"),

        keysList = List.Zip({types, objCodes, revisions}),

        keys = List.Transform(keysList, each Text.Combine(_)),

        data = if (List.Count(connectionsList) > 0)
            then
                List.Transform(objCodes, each GetObjectConnectionsNavTableRecursive(address, _))
                //List.Transform(objCodes, each GetObjectConnectionsNavTable(address, _))
            else
                response,

        itemKind = if (List.Count(connectionsList) > 0)
            then
                List.Transform(labels, each "Folder")
            else
                List.Transform(labels, each "Table"),

        itemName = if (List.Count(connectionsList) > 0)
            then
                List.Transform(labels, each "Folder")
            else
                List.Transform(labels, each "Table"),

        isLeaf = if (List.Count(connectionsList) > 0)
            then
                List.Transform(labels, each false)
            else
                List.Transform(labels, each true),
        
        
        // NavTable for the recursive connections
        subLevelTableHeader = {"Name", "Key", "Data", "ItemKind", "ItemName", "IsLeaf"},
        subLevelTable = Table.FromColumns({labels, keys, data, itemKind, itemName, isLeaf}, subLevelTableHeader),
        
        subLevelNavTable = Table.ToNavigationTable(subLevelTable, {"Key"}, "Name", "Data", "ItemKind", "ItemName", "IsLeaf"),

        secondLevelConnectionsLabel    = "Hierarchy",
        secondLevelConnectionsKey      = "connectionsRecursive",
        secondLevelConnectionsData     = subLevelNavTable,
        secondLevelConnectionsItemKind = "Cube",
        secondLevelConnectionsItemName = "Cube",
        secondLevelConnectionsIsLeaf   = false,

        

        allLabels   = {objectLabel,    firstLevelConnectionsLabel   , secondLevelConnectionsLabel},
        allKeys     = {objectKey,      firstLevelConnectionsKey     , secondLevelConnectionsKey},
        allData     = {objectData,     firstLevelConnectionsData    , secondLevelConnectionsData},
        allItemKind = {objectItemKind, firstLevelConnectionsItemKind, secondLevelConnectionsItemKind},
        allItemName = {objectItemName, firstLevelConnectionsItemName, secondLevelConnectionsItemName},
        allIsLeaf   = {objectIsLeaf,   firstLevelConnectionsIsLeaf  , secondLevelConnectionsIsLeaf},
        
        // NavTable for the level that shows the object itself, the object's connections flat and the object's connections recursively
        topLevelTableHeader = {"Name", "Key", "Data", "ItemKind", "ItemName", "IsLeaf"},
        topLevelTable = Table.FromColumns({allLabels, allKeys, allData, allItemKind, allItemName, allIsLeaf}, topLevelTableHeader),

        NavTable = Table.ToNavigationTable(topLevelTable, {"Key"}, "Name", "Data", "ItemKind", "ItemName", "IsLeaf")
    in
        NavTable;

GetObjectConnectionsRecursiveHelper = (
    address as text,
    objCode as text,
    hierarchyLevels as number,
    removeDuplicates as logical) as table =>
    let
        table = GetObjectConnectionsRecursive(
            address,
            objCode,
            objCode,
            hierarchyLevels,
            hierarchyLevels,
            removeDuplicates),

        handleDuplicatesTable = if (removeDuplicates = true)
            then
                Table.Distinct(table)
            else
                table

        in
            handleDuplicatesTable;


GetObjectConnectionsRecursive = (
    address as text,
    currentObjCode as text,
    originalObjCode as text,
    currentHierarchyLevel as number,
    originalHierarchyLevel as number,
    removeDuplicates as logical) as table =>
    let
        currentObject = ObjectsGetSingle(address, currentObjCode),
        currentObjectTable = Table.FromRecords({currentObject}),

        table = if (currentHierarchyLevel > 0)
            then
                Table.Combine(
                        List.Combine(
                        {
                            {
                                currentObjectTable
                            }
                            ,
                            // It would actually be wiser to not call ObjectsGetSingle for all the connections
                            // unless we need their connections as well, since the parent object has a list of the connections
                            // in the parent ObjectsGetSingle call already

                            if (currentHierarchyLevel - 1 = 0)
                                then
                                    {Table.FromRecords(currentObject[connections])}
                                else
                                    List.Transform(
                                        currentObject[connections],
                                        each if (_[objCode] = originalObjCode) // Make sure the recursion does not happen for the original object code again
                                            then
                                                #table(Table.ColumnNames(currentObjectTable),{})
                                            else
                                                GetObjectConnectionsRecursive(
                                                    address,
                                                    _[objCode],
                                                    originalObjCode,
                                                    currentHierarchyLevel-1,
                                                    originalHierarchyLevel,
                                                    removeDuplicates)
                                    )
                        }
                        )
                )
            else
                if (currentHierarchyLevel = originalHierarchyLevel)
                    then
                        currentObjectTable
                    else
                        #table(Table.ColumnNames(currentObjectTable), {}),

        distinctTable = Table.Distinct(table)
    in
        distinctTable;


GetClassifications = (address as text) as table =>
    let
        response = DirectAPICall(address, "objects/classifications"),

        table = Table.FromRecords(response[data])

    in
        table;

GetFileListHandlerType = type function (
    baseClass as (type text meta [
        Documentation.FieldCaption      = "Base class",
        Documentation.FieldDescription  = "Base class of the object",
        Documentation.SampleValues      = {"doc"}
        ]),
    objType as (type text meta [
        Documentation.FieldCaption      = "Object type",
        Documentation.FieldDescription  = "Type of the object",
        Documentation.SampleValues      = {"model"}
        ]),
    objCode as (type text meta [
        Documentation.FieldCaption      = "Object code",
        Documentation.FieldDescription  = "Object code of the object",
        Documentation.SampleValues      = {"VX3000016"}
        ]),
    revision as (type text meta [
        Documentation.FieldCaption      = "Revision",
        Documentation.FieldDescription  = "Revision of the object",
        Documentation.SampleValues      = {"0"}
        ])
    )
    as table meta [];

GetFileList = (
    address as text,
    baseClass as text,
    objType as text,
    objCode as text,
    revision as text
    ) as table =>
    let
        table = GetTableThroughAuthorization(
            address,
            "files/list",
            "{ " &
                """baseClass"": """  & baseClass & """," &
                """type"": """       & objType   & """," &
                """objCode"": """    & objCode   & """," &
                """revision"": """   & revision  & """"  &
            " }")
    in
        table;

/************************************
* Users
************************************/

GetUsersList = (address as text) as any =>
    let
        table = GetTableThroughAuthorization(
            address,
            "users/list")
    in
        table;

// Data Source Kind description
VertexFlow = [

    // Test connection for On-Premises Data Gateway
    TestConnection = (dataSourcePath) =>
        let
            json = Json.Document(dataSourcePath),
            address = json[address]
        in
            {"VertexFlow.Contents", address},

    Authentication = [
        UsernamePassword = [] // Flow username and password, stored in Power BI
    ],

    Label = "Vertex Flow" // Label text shown in the login popup in Power BI
    
];

// Data Source UI publishing description
VertexFlow.Publish = [
    Beta = false,
    Category = "Online",
    //Description = "Vertex Flow",
    //Label = "Vertex Flow",
    ButtonText = { Extension.LoadString("ButtonTitle"), Extension.LoadString("ButtonHelp") },
    LearnMoreUrl = "https://kb.vertex.fi/flow2022fi/flow-systeeminhoito/liitynnaet-muihin-jaerjestelmiin/vertex-flow-openapi",
    SourceImage = VertexFlow.Icons,
    SourceTypeImage = VertexFlow.Icons
];

VertexFlow.Icons = [
    Icon16 = { Extension.Contents("VertexFlowConnector16.png"), Extension.Contents("VertexFlowConnector20.png"), Extension.Contents("VertexFlowConnector24.png"), Extension.Contents("VertexFlowConnector32.png") },
    Icon32 = { Extension.Contents("VertexFlowConnector32.png"), Extension.Contents("VertexFlowConnector40.png"), Extension.Contents("VertexFlowConnector48.png"), Extension.Contents("VertexFlowConnector64.png") }
];

// Helper function from Microsoft
Table.ToNavigationTable = (
    table as table,
    keyColumns as list,
    nameColumn as text,
    dataColumn as text,
    itemKindColumn as text,
    itemNameColumn as text,
    isLeafColumn as text
) as table =>
    let
        tableType = Value.Type(table),
        newTableType = Type.AddTableKey(tableType, keyColumns, true) meta 
        [
            NavigationTable.NameColumn = nameColumn, 
            NavigationTable.DataColumn = dataColumn,
            NavigationTable.ItemKindColumn = itemKindColumn, 
            Preview.DelayColumn = itemNameColumn, 
            NavigationTable.IsLeafColumn = isLeafColumn
        ],
        navigationTable = Value.ReplaceType(table, newTableType)
    in
        navigationTable;

// Helper function from Microsoft with modifications from GitHub thread
Table.ToNavigationTableNoDelay = (
    table as table,
    keyColumns as list,
    nameColumn as text,
    dataColumn as text,
    itemKindColumn as text,
    itemNameColumn as text,
    isLeafColumn as text
) as table =>
    let
        tableType = Value.Type(table),
        newTableType = Type.AddTableKey(tableType, keyColumns, true) meta 
        [
            NavigationTable.NameColumn = nameColumn, 
            NavigationTable.DataColumn = dataColumn,
            NavigationTable.ItemKindColumn = itemKindColumn,
            NavigationTable.ItemNameColumn = itemNameColumn, // ?
            //Preview.DelayColumn = itemNameColumn, 
            NavigationTable.IsLeafColumn = isLeafColumn
        ],
        navigationTable = Value.ReplaceType(table, newTableType)
    in
        navigationTable;
